/* PID-control for thermostat */

#include <inttypes.h>
#include "thermo.h"

/* digital filter on temperature
   designed with 
   http://www.dsptutor.freeuk.com/IIRFilterDesign/IIRFilterDesign.html
   Coefficients for 10HZ bei 8000Hz sample rate
*/
#define A0 1.5336025E-5
#define A1 3.067205E-5
#define A2 1.5336025E-5
#define B0 1.0
#define B1 -1.9888929
#define B2 0.98895425

/* updates current temperature every second */
void temp_filter(uint8_t init) {
  float temp = TEMP(cur_temp);

  static float
    x0, x1, x2,
    y0, y1, y2;

  if(init) {
    x2 = x1 = x0 = temp;
    y2 = y1 = y0 = temp;
  } else {
    x2 = x1; x1 = x0; x0 = temp;
    y2 = y1; y1 = y0;
  }

  y0 = (A0 * x0 + A1 * x1 + A2 * x2
	- (B1 * y1 + B2 * y2)) / B0;

  temperature = (int16_t)y0;
}

/* PID (or actually, in the current parametration, PI) controller
   setpoint + current value from global vars 
*/
uint8_t pid(uint8_t init) {
  static float l_out, out;
  float err;
  static int16_t l_ist, ll_ist;

  err = setpoint - temperature;

  if(init) {
    l_out = pid_P * err + CTL_MIN;
    l_ist = temperature;
    ll_ist = temperature;
  }

  out = l_out - pid_P * ((temperature - l_ist)
			 - pid_I * err
			 + pid_D * (temperature - 2 * l_ist + ll_ist)
			 );
  /* velocity has its windup problems, too */
  if((err < 0) && (out > l_out))
    out = l_out;

  if(out > (float)CTL_MAX) {
    out = (float)CTL_MAX;
  }
  if(out < (float)CTL_MIN) {
    out = (float)CTL_MIN;
  }

  l_out = out;
  ll_ist = l_ist;
  l_ist = temperature;
  return (uint8_t)out;
}

#if 0

uint8_t pid(int soll, int ist, uint8_t init) {
  static float
    /* PID */
    l_out,
    out,
    /* filter */
    x0, x1, x2,
    y0, y1, y2;
  float err;

  if(init) {
    x2 = x1 = x0 = ist;
    y2 = y1 = y0 = ist;
  } else {
    x2 = x1; x1 = x0; x0 = ist;
    y2 = y1; y1 = y0;
  }
  y0 = (A0 * x0 + A1 * x1 + A2 * x2
	- (B1 * y1 + B2 * y2)) / B0;

  err = (float)soll - y0;

  if(init) {
    l_out = pid_P * err + CTL_MIN;
  }

  out = l_out - pid_P * ((y0 - y1)
			 - pid_I * err
			 + pid_D * (y0 - 2 * y1 + y2)
			 );
  /* velocity has its windup problems, too */
  if((err < 0) && (out > l_out))
    out = l_out;

  if(out > (float)CTL_MAX) {
    out = (float)CTL_MAX;
  }
  if(out < (float)CTL_MIN) {
    out = (float)CTL_MIN;
  }
  
  l_out = out;
  temperature = (uint16_t)y0;
  return (uint8_t)out;
}
#endif
